#![no_std]
use core::cell::UnsafeCell;
// use core::marker::PhantomData;
use core::ops::{Deref, DerefMut, Drop};
use cortex_m::interrupt;

/// Internal replacement for `static mut T`
// TODO: Decide name and location.
pub struct RacyCell<T>(UnsafeCell<T>);

impl<T> RacyCell<T> {
    /// Create a RacyCell
    pub const fn new(value: T) -> Self {
        RacyCell(UnsafeCell::new(value))
    }

    /// Get &mut T
    pub unsafe fn get_mut_unchecked(&self) -> &mut T {
        &mut *self.0.get()
    }
}

// The type wrapped need to be Sync for RacyCell<T> to be Sync
//unsafe impl<T> Sync for RacyCell<T> where T: Sync {}
unsafe impl<T> Sync for RacyCell<T> {}

#[inline]
pub fn interrupt_free<F, R>(f: F) -> R
where
    F: FnOnce() -> R,
{
    if cfg!(target_arch = "arm") {
        interrupt::free(|_| f())
    } else {
        f()
    }
}

#[derive(Debug)]
pub enum Error {
    Uninitialized,
    OutOfMemory,
    AlreadyInitialized,
}

trait Pool<T>
where
    T: Send + Sync,
{
    fn dealloc(&'static self, index: usize);
    fn index_deref(&'static self, index: usize) -> &T;
    fn index_deref_mut(&'static self, index: usize) -> &mut T;
}

pub struct B<T: 'static + Sized + Send + Sync> {
    index: usize,
    allocator: &'static dyn Pool<T>,
}

impl<T: 'static + Sized + Send + Sync> Drop for B<T> {
    fn drop(&mut self) {
        self.allocator.dealloc(self.index)
    }
}

impl<T: 'static + Sized + Send + Sync> Deref for B<T> {
    type Target = T;
    fn deref(&self) -> &T {
        self.allocator.index_deref(self.index)
    }
}

impl<T: 'static + Sized + Send + Sync> DerefMut for B<T> {
    fn deref_mut(&mut self) -> &mut T {
        self.allocator.index_deref_mut(self.index)
    }
}

// pub struct Box<T: 'static + Sized, const S: usize> {
//     index: usize,
//     allocator: &'static PoolMan<T, S>,
// }

// impl<T: 'static, const S: usize> Drop for Box<T, S> {
//     fn drop(&mut self) {
//         self.allocator.dealloc(self.index)
//     }
// }

// impl<T: 'static, const S: usize> Deref for Box<T, S> {
//     type Target = T;
//     fn deref(&self) -> &T {
//         unsafe {
//             self.allocator
//                 .data /* RacyCell */
//                 .get_mut_unchecked() /* Array */
//                 .get_unchecked(self.index)
//         }
//     }
// }

// impl<T: 'static, const S: usize> DerefMut for Box<T, S> {
//     fn deref_mut(&mut self) -> &mut T {
//         unsafe {
//             self.allocator
//                 .data /* RacyCell */
//                 .get_mut_unchecked() /* Array */
//                 .get_unchecked_mut(self.index)
//         }
//     }
// }

pub struct PoolMan<T: 'static, const S: usize> {
    data: RacyCell<[T; S]>,
    free: RacyCell<[usize; S]>,
    head: RacyCell<usize>,
    init: RacyCell<bool>,
}



impl<T, const S: usize> PoolMan<T, S>
where
    T: Sized + Send + Sync,
{
    pub const fn new(v: [T; S]) -> Self
    where
        T: Sized + Send + Sync,
    {
        Self {
            data: RacyCell::new(v),
            free: RacyCell::new([0; S]),
            head: RacyCell::new(0),
            init: RacyCell::new(false),
        }
    }

    #[inline]
    pub fn init(&self) -> Result<(), Error> {
        interrupt_free(|| {
            if *unsafe { self.init.get_mut_unchecked() } {
                Err(Error::AlreadyInitialized)
            } else {
                let free = unsafe { self.free.get_mut_unchecked() };
                for (index, value) in free.iter_mut().enumerate() {
                    *value = index + 1;
                }
                *unsafe { self.init.get_mut_unchecked() } = true;
                Ok(())
            }
        })
    }

    // // Safety:
    // //  dealloc can only be called by dropping Box
    // //  which can happen only if already initialized
    // //  thus it is safe to assume init == true
    // //
    // // not accessible from user code
    // #[inline]
    // fn dealloc(&self, index: usize) {
    //     interrupt_free(|| unsafe {
    //         *self.free.get_mut_unchecked().get_unchecked_mut(index) =
    //             *self.head.get_mut_unchecked();
    //         *self.head.get_mut_unchecked() = index;
    //     });
    // }

    // #[inline]
    // pub fn alloc(&'static self) -> Result<Box<T, S>, Error> {
    //     // this check does not need to be inside of critical section
    //     // as initialization is monotonic, cannot be undone
    //     if !unsafe { *self.init.get_mut_unchecked() } {
    //         Err(Error::Uninitialized)
    //     } else {
    //         interrupt_free(|| unsafe {
    //             let index = *self.head.get_mut_unchecked();
    //             *self.head.get_mut_unchecked() =
    //                 *self.free.get_mut_unchecked().get_unchecked(index);
    //             if *self.head.get_mut_unchecked() > S {
    //                 Err(Error::OutOfMemory)
    //             } else {
    //                 Ok(Box {
    //                     index,
    //                     allocator: self,
    //                 })
    //             }
    //         })
    //     }
    // }

    pub fn allocb(&'static self) -> Result<B<T>, Error> {
        // this check does not need to be inside of critical section
        // as initialization is monotonic, cannot be undone
        if !unsafe { *self.init.get_mut_unchecked() } {
            Err(Error::Uninitialized)
        } else {
            interrupt_free(|| unsafe {
                let index = *self.head.get_mut_unchecked();
                *self.head.get_mut_unchecked() =
                    *self.free.get_mut_unchecked().get_unchecked(index);
                if *self.head.get_mut_unchecked() > S {
                    Err(Error::OutOfMemory)
                } else {
                    Ok(B {
                        index,
                        allocator: self,
                    })
                }
            })
        }
    }
}

// sealed, should only be externally accessible
impl<T, const S: usize> Pool<T> for PoolMan<T, S>
where
    T: Sized + Send + Sync,
{
    #[inline]
    fn dealloc(&'static self, index: usize) {
        interrupt_free(|| unsafe {
            *self.free.get_mut_unchecked().get_unchecked_mut(index) =
                *self.head.get_mut_unchecked();
            *self.head.get_mut_unchecked() = index;
        });
    }

    fn index_deref(&'static self, index: usize) -> &'static T {
        unsafe { self.data.get_mut_unchecked().get_unchecked(index) }
    }

    fn index_deref_mut(&'static self, index: usize) -> &'static mut T {
        unsafe { self.data.get_mut_unchecked().get_unchecked_mut(index) }
    }
}
